from datetime import timezone
from functools import partial, update_wrapper

from django.utils.cache import get_conditional_response
from django.utils.http import http_date, quote_etag
from rest_framework import status
from rest_framework.metadata import BaseMetadata
from rest_framework.response import Response
from rest_framework.views import APIView

from scrud_django import __version__
from scrud_django.utils import get_string_or_evaluate, link_content


class ScrudfulViewFunc:
    def __init__(
        self,
        view_method,
        view_is_class_method=True,
        etag_func=None,
        last_modified_func=None,
        schema_link_or_func=None,
        schema_rel_or_func=None,
        schema_type_or_func=None,
        context_link_or_func=None,
        context_rel_or_func=None,
        context_type_or_func=None,
        http_schema_or_func=None,
    ):
        self.view_is_class_method = view_is_class_method
        self.view_method = view_method
        self.etag_func = etag_func
        self.last_modified_func = last_modified_func
        self.schema_link_or_func = schema_link_or_func
        self.schema_rel_or_func = schema_rel_or_func
        self.schema_type_or_func = schema_type_or_func
        self.context_link_or_func = context_link_or_func
        self.context_rel_or_func = context_rel_or_func
        self.context_type_or_func = context_type_or_func
        self.http_schema_or_func = http_schema_or_func
        update_wrapper(self, self.view_method)

    def __get__(self, obj, objtype):
        return partial(self.__call__, obj)

    def __call__(self, *args, **kwargs):
        if self.view_is_class_method:
            request = args[1]
        else:
            request = args[0]
        if request.method in ("PUT", "DELETE"):
            missing_required_headers = []
            if self.etag_func and not request.META.get("HTTP_IF_MATCH"):
                missing_required_headers.append("If-Match")
            if self.last_modified_func and not request.META.get(
                "HTTP_IF_UNMODIFIED_SINCE"
            ):
                missing_required_headers.append("If-Unmodified-Since")
            if missing_required_headers:
                # TODO Define standard error json
                response = Response(
                    {"missing-required-headers": missing_required_headers}, status=400,
                )
                return response

        # Compute values (if any) for the requested resource.
        def get_last_modified():
            if self.last_modified_func:
                last_modified = self.last_modified_func(*args, **kwargs)
                if last_modified:
                    return http_date(
                        last_modified.replace(tzinfo=timezone.utc).timestamp()
                    )
            return None

        etag = None
        last_modified = None
        if request.method not in ("POST", "OPTIONS"):
            if self.etag_func:
                etag = self.etag_func(*args, **kwargs)
                etag = etag + __version__ if etag else None
            etag = quote_etag(etag) if etag else None
            last_modified = get_last_modified()
        else:
            etag = None
            last_modified = None
        response = get_conditional_response(
            request, etag=etag, last_modified=last_modified
        )
        if response is None:
            response = self.view_method(*args, **kwargs)
            schema_link = self.schema_link_header(*args, **kwargs) or ""
            context_link = self.context_link_header(*args, **kwargs) or ""
            join_links = ", " if schema_link and context_link else ""
            link_content = schema_link + join_links + context_link
            if etag:
                response["ETag"] = etag
            if last_modified:
                response["Last-Modified"] = last_modified
            if link_content:
                response["Link"] = link_content
            self.add_expose_headers(response)
        return response

    def add_expose_headers(self, response):
        """If the Link and/or Location header are provided on the response add the
        'Access-Control-Expose-Headers` header to expose them over CORS requests.
        """
        expose_headers = ""
        if "Link" in response:
            expose_headers = "Link"
        if "Location" in response:
            if expose_headers:
                expose_headers = expose_headers + ", "
            expose_headers = expose_headers + "Location"
        if expose_headers:
            response["Access-Control-Expose-Headers"] = expose_headers

    def schema_link(self, *args, **kwargs):
        return get_string_or_evaluate(self.schema_link_or_func, *args, **kwargs)

    def schema_link_header(self, *args, **kwargs):
        link = self.schema_link(*args, **kwargs)
        if link:
            link_rel = (
                get_string_or_evaluate(self.schema_rel_or_func, *args, **kwargs,)
                or "describedBy"
            )
            link_type = (
                get_string_or_evaluate(self.schema_type_or_func, *args, **kwargs,)
                or "application/json"
            )
            return link_content(link, link_rel, link_type)
        return None

    def context_link(self, *args, **kwargs):
        return get_string_or_evaluate(self.context_link_or_func, *args, **kwargs)

    def context_link_header(self, *args, **kwargs):
        link = self.context_link(*args, **kwargs)
        if link:
            link_rel = (
                get_string_or_evaluate(self.context_rel_or_func, *args, **kwargs,)
                or "http://www.w3.org/ns/json-ld#context"
            )
            link_type = (
                get_string_or_evaluate(self.context_type_or_func, *args, **kwargs,)
                or "application/ld+json"
            )
            return link_content(link, link_rel, link_type)
        return None


def scrudful(
    etag_func=None,
    last_modified_func=None,
    schema_link_or_func=None,
    schema_rel_or_func=None,
    schema_type_or_func=None,
    context_link_or_func=None,
    context_rel_or_func=None,
    context_type_or_func=None,
    http_schema_or_func=None,
):
    """Decorator to make a view method SCRUDful"""
    # TODO what about 400 Bad Request context and schema?
    def decorator(view_method):
        return ScrudfulViewFunc(
            view_method,
            etag_func=etag_func,
            last_modified_func=last_modified_func,
            schema_link_or_func=schema_link_or_func,
            schema_rel_or_func=schema_rel_or_func,
            schema_type_or_func=schema_type_or_func,
            context_link_or_func=context_link_or_func,
            context_rel_or_func=schema_rel_or_func,
            context_type_or_func=schema_type_or_func,
            http_schema_or_func=http_schema_or_func,
        )

    return decorator


def scrudful_api_view(
    etag_func=None,
    last_modified_func=None,
    schema_link_or_func=None,
    schema_rel_or_func=None,
    schema_type_or_func=None,
    context_link_or_func=None,
    context_rel_or_func=None,
    context_type_or_func=None,
    http_schema_or_func=['GET'],
):
    def decorator(view_method, *args, **kwargs):

        http_method_names = http_schema_or_func
        allowed_methods = set(http_method_names) | {'options'}

        cls_attr = {
            '__doc__': view_method.__doc__,
            'metadata_class': ScrudfulAPIViewMetadata,
        }

        handler = ScrudfulViewFunc(
            lambda self, *args, **kwargs: view_method(*args, **kwargs),
            etag_func=etag_func,
            last_modified_func=last_modified_func,
            schema_link_or_func=schema_link_or_func,
            schema_rel_or_func=schema_rel_or_func,
            schema_type_or_func=schema_type_or_func,
            context_link_or_func=context_link_or_func,
            context_rel_or_func=context_rel_or_func,
            context_type_or_func=context_type_or_func,
        )

        for method in http_method_names:
            cls_attr[method.lower()] = handler

        ScrudAPIView = type('ScrudAPIView', (APIView,), cls_attr)

        ScrudAPIView.http_method_names = [method.lower() for method in allowed_methods]

        ScrudAPIView.__name__ = view_method.__name__
        ScrudAPIView.__module__ = view_method.__module__

        ScrudAPIView.permission_classes = getattr(
            view_method, 'permission_classes', APIView.permission_classes
        )
        ScrudAPIView.schema = getattr(view_method, 'schema', APIView.schema)
        ScrudAPIView.schema_link_or_func = schema_link_or_func
        ScrudAPIView.context_link_or_func = context_link_or_func

        # ScrudAPIView.options = options

        new_view_method = ScrudAPIView.as_view()

        return new_view_method

    return decorator


class ScrudfulMetadata(BaseMetadata):
    def determine_metadata(self, request, view, *args, **kwargs):
        if len(args) > 0 or len(kwargs) > 0:  # this is a detail request
            return self.determine_metadata_for_detail(request, view)
        return self.determine_metadata_for_list(request, view)

    def determine_metadata_for_detail(self, request, view):
        metadata = dict()
        metadata.update(
            {
                key: value
                for key, value in {
                    "get": self.determine_metadata_for_get(request, view, "retrieve"),
                    "put": self.determine_metadata_for_put(request, view),
                    "delete": self.determine_metadata_for_delete(request, view),
                }.items()
                if value is not None
            }
        )
        return metadata

    def determine_metadata_for_list(self, request, view):
        metadata = dict()
        metadata.update(
            {
                key: value
                for key, value in {
                    "post": self.determine_metadata_for_post(request, view),
                    "get": self.determine_metadata_for_get(request, view, "list"),
                }.items()
                if value is not None
            }
        )
        return metadata

    def get_method(self, view, name):
        method_partial = getattr(view, name, None)
        if method_partial:
            return method_partial.func.__self__
        return None

    def determine_metadata_for_post(self, request, view, name="create"):
        create_method = self.get_method(view, name)
        if create_method is None:
            return None
        schema_link = create_method.schema_link(view, request)
        context_link = create_method.context_link(view, request)
        request_body = {
            "description": "The content for the resource to be created.",
            "required": True,
        }
        if schema_link or context_link:
            json_content = {}
            if schema_link:
                json_content["schema"] = schema_link
            if context_link:
                json_content["context"] = context_link
            request_body["content"] = {
                "application/json": json_content,
            }
        metadata = {
            "requestBody": request_body,
            "responses": {"201": {"description": "CREATED"}},
        }
        return metadata

    def determine_metadata_for_get(self, request, view, name):
        list_method = self.get_method(view, name)
        if list_method is None:
            return
        schema_link = list_method.schema_link(view, request)
        context_link = list_method.context_link(view, request)
        json_content = None
        if schema_link or context_link:
            json_content = {}
            if schema_link:
                json_content["schema"] = schema_link
            if context_link:
                json_content["context"] = context_link
        responses = {
            "200": {"description": "OK"},
        }
        if json_content:
            responses["200"]["content"] = {
                "application/json": json_content,
            }
        return {
            "responses": responses,
        }

    def required_conditional_headers(self, method):
        supports_etag = method.etag_func is not None
        supports_last_modified = method.last_modified_func is not None
        parameters = None
        if supports_etag or supports_last_modified:
            parameters = []
            if supports_etag:
                parameters.append(
                    {
                        "in": "header",
                        "name": "If-Match",
                        "schema": {"type": "string"},
                        "required": True,
                    }
                )
            if supports_last_modified:
                parameters.append(
                    {
                        "in": "header",
                        "name": "If-Unmodified-Since",
                        "schema": {"type": "string"},
                        "required": True,
                    }
                )
        return parameters

    def determine_metadata_for_put(self, request, view, name="update"):
        update_method = self.get_method(view, name)
        if update_method is None:
            return
        schema_link = update_method.schema_link(view, request)
        context_link = update_method.context_link(view, request)
        request_body = {
            "description": "The content for the resource to be created.",
            "required": True,
        }
        if schema_link or context_link:
            json_content = {}
            if schema_link:
                json_content["schema"] = schema_link
            if context_link:
                json_content["context"] = context_link
            request_body["content"] = {
                "application/json": json_content,
            }
        metadata = {
            "requestBody": request_body,
            "responses": {"200": {"description": "OK"}},
        }
        parameters = self.required_conditional_headers(update_method)
        if parameters:
            metadata["parameters"] = parameters
        return metadata

    def determine_metadata_for_delete(self, request, view, name="destroy"):
        delete_method = self.get_method(view, name)
        if delete_method is None:
            return None
        metadata = {
            "responses": {"200": {"description": "OK"}},
        }
        parameters = self.required_conditional_headers(delete_method)
        if parameters:
            metadata["parameters"] = parameters
        return metadata


class ScrudfulAPIViewMetadata(ScrudfulMetadata):
    def determine_metadata(self, request, view, *args, **kwargs):
        metadata = dict()
        metadata.update(
            {
                key: value
                for key, value in {
                    "get": self.determine_metadata_for_get(request, view, "get"),
                    "post": self.determine_metadata_for_post(request, view, "post"),
                    "put": self.determine_metadata_for_put(request, view, "put"),
                    "delete": self.determine_metadata_for_delete(
                        request, view, "delete"
                    ),
                }.items()
                if value is not None
            }
        )
        return metadata


def options(view_instance, request, *args, **kwargs):
    data = ScrudfulMetadata().determine_metadata(
        request, view_instance, *args, **kwargs
    )
    return Response(data, status=status.HTTP_200_OK)


def scrudful_viewset(cls):
    setattr(cls, "options", options)
    meta = getattr(cls, "Meta", None)
    etag_func = getattr(meta, "etag_func", None)
    last_modified_func = getattr(meta, "last_modified_func", None)
    schema_link_or_func = getattr(meta, "schema_link_or_func", None)
    schema_rel_or_func = getattr(meta, "schema_rel_or_func", None)
    schema_type_or_func = getattr(meta, "schema_type_or_func", None)
    context_link_or_func = getattr(meta, "context_link_or_func", None)
    context_rel_or_func = getattr(meta, "context_rel_or_func", None)
    context_type_or_func = getattr(meta, "context_type_or_func", None)
    scrudful_item = scrudful(
        etag_func=etag_func,
        last_modified_func=last_modified_func,
        schema_link_or_func=schema_link_or_func,
        schema_rel_or_func=schema_rel_or_func,
        schema_type_or_func=schema_type_or_func,
        context_link_or_func=context_link_or_func,
        context_rel_or_func=context_rel_or_func,
        context_type_or_func=context_type_or_func,
    )
    for method_name in ("create", "retrieve", "update", "destroy"):
        method = getattr(cls, method_name, None)
        setattr(cls, method_name, scrudful_item(method))
    if hasattr(cls, "list"):
        scrudful_list = scrudful(
            etag_func=getattr(meta, "list_etag_func", None),
            last_modified_func=getattr(meta, "list_last_modified_func", None),
            schema_link_or_func=getattr(meta, "list_schema_link_or_func", None),
            schema_rel_or_func=getattr(meta, "list_schema_rel_or_func", None),
            schema_type_or_func=getattr(meta, "list_schema_type_or_func", None),
            context_link_or_func=getattr(meta, "list_context_link_or_func", None),
            context_rel_or_func=getattr(meta, "list_context_rel_or_func", None),
            context_type_or_func=getattr(meta, "list_context_type_or_func", None),
        )
        list_method = getattr(cls, "list")
        setattr(cls, "list", scrudful_list(list_method))
    return cls



# Generated by Django 3.1 on 2020-09-12 15:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('scrud_django', '0002_auto_20200826_0345'),
    ]

    operations = [
        migrations.AddField(
            model_name='resourcetype',
            name='revision',
            field=models.CharField(default='1', max_length=40),
            preserve_default=False,
        ),
    ]



def link_content(url, rel, content_type):
    return f"<{url}>; rel=\"{rel}\"; type=\"{content_type}\""


def get_string_or_evaluate(string_or_func, *args, **kwargs):
    if not string_or_func:
        return None
    if isinstance(string_or_func, str):
        return string_or_func
    return string_or_func(*args, **kwargs)



"""
SCRUD-Django - Semantic Create, Read, Update and Delete Django application.
"""

__version__ = "0.1.9"
import base64
import datetime

from django.utils.timezone import datetime as ddt


class ScrudServices:
    last_modified: datetime.datetime
    services: dict
    etag: str

    def __init__(self):
        self.update_last_modified()
        print(F"Initializing services - last modified = {self.last_modified}")
        self.services = {}

    def _set_etag(self):
        self.etag = base64.b64encode(
            self.last_modified.isoformat().encode('utf-8')
        ).decode('utf-8')

    def update_last_modified(self):
        self.last_modified = ddt.now()
        self._set_etag()

    def add_service(self, key, value):
        self.services[key] = value
        self.update_last_modified()

    def get_last_modified(self):
        return self.last_modified

    def get_etag(self):
        return self.etag


services = ScrudServices()



import json
from datetime import timezone

from django.db.models import Manager
from django.utils.http import http_date, quote_etag
from rest_framework import serializers
from rest_framework.reverse import reverse

from scrud_django import models
from scrud_django.headers import get_link_header_for
from scrud_django.models import Resource
from scrud_django.registration import ResourceRegistration
from scrud_django.validators import JsonValidator


class EnvelopeSerializer(serializers.ListSerializer):
    def to_representation(self, data):
        iterable = data.all() if isinstance(data, Manager) else data
        return [self.child.to_representation(item, envelope=True) for item in iterable]


class ResourceSerializer(serializers.Serializer):
    class Meta:
        model = models.Resource
        list_serializer_class = EnvelopeSerializer
        validators = [JsonValidator()]

    def __init__(self, *args, **kwargs):
        self.resource_type = kwargs.pop("resource_type")
        super().__init__(*args, **kwargs)

    def to_representation(self, instance, envelope=False, context=None):
        if type(instance.content) is str:
            content = json.loads(instance.content)
        else:
            content = instance.content
        if not envelope:
            return content
        request = self._context["request"]
        last_modified = http_date(
            instance.modified_at.replace(tzinfo=timezone.utc).timestamp()
        )
        return {
            'href': reverse(
                self.resource_type.route_name_detail(),
                args=[instance.id],
                request=request,
            ),
            'last_modified': last_modified,
            'etag': quote_etag(instance.etag),
            'link': get_link_header_for(self.resource_type, request),
            'content': content,
        }

    def to_internal_value(self, data):
        return data

    def create(self, validated_data):
        instance = Resource(content=validated_data, resource_type=self.resource_type)
        return ResourceRegistration.register_instance(instance)

    def update(self, instance, validated_data):
        return ResourceRegistration.update(
            content=validated_data,
            register_type=self.resource_type.slug,
            slug=instance.slug,
        )


class JSONSchemaSerializer(serializers.Serializer):
    """Serializer for JSON-Schema data."""

    class Meta:
        fields = ["$id", "$schema", "title", "description", "properties"]


class JSONLDSerializer(serializers.Serializer):
    """Serializer for JSON-LD data."""

    class Meta:
        fields = ["$id", "$schema", "title", "description", "properties"]



import json
import logging
from datetime import datetime
from uuid import uuid4

from django.conf import settings
from django.db import transaction
from django.db.utils import OperationalError, ProgrammingError
from django.shortcuts import get_object_or_404
from django.urls import path

from scrud_django import services
from scrud_django.models import Resource, ResourceType


class ResourceTypeRegistration:
    def __init__(
        self,
        resource_type_uri: str,
        revision: str,
        schema_resource: Resource = None,
        schema_func: str = None,
        schema_uri: str = None,
        context_resource: Resource = None,
        context_func: str = None,
        context_uri: str = None,
        slug: str = None,
    ):
        """
        Register resource type model.
        """
        search = ResourceType.objects.filter(type_uri=resource_type_uri)
        if search:
            self.resource_type = search[0]
        else:
            resource_type = ResourceType(
                type_uri=resource_type_uri,
                schema_uri=schema_uri,
                context_uri=context_uri,
                slug=slug,
                etag=uuid4().hex,
                modified_at=datetime.now(),
            )
            self.resource_type = resource_type

        if self.resource_type.revision != revision or settings.DEBUG:
            self.resource_type.revision = revision
            if schema_resource is None and schema_func:
                self.resource_type.schema = register_json_schema(
                    schema_func, self.resource_type.schema
                )
            if context_resource is None and context_func:
                self.resource_type.context = register_json_ld_context(
                    context_func, self.resource_type.context
                )
            self.resource_type.save()

        # Register with service catalog
        services.add_service(self.resource_type.type_uri, self.resource_type.slug)
        self.register_urls()

    def register_urls(self):
        """Register resource type url."""
        from scrud_django import views

        scrud_list = views.ResourceViewSet.as_view(
            {'get': 'list', 'post': 'create'},
            resource_type_name=self.resource_type.slug,
        )
        scrud_detail = views.ResourceViewSet.as_view(
            {'get': 'retrieve', 'put': 'update', 'delete': 'destroy'},
            resource_type_name=self.resource_type.slug,
        )
        self.urls = [
            path(
                f'{self.resource_type.slug}/',
                scrud_list,
                # name=f'{resource_type.slug}-list',
                name=self.resource_type.route_name_list(),
            ),
            path(
                f'{self.resource_type.slug}/<str:slug>/',
                scrud_detail,
                # name=f'{resource_type.slug}-detail',
                name=self.resource_type.route_name_detail(),
            ),
        ]


class ResourceRegistration:
    @staticmethod
    def register(content: str, register_type: str) -> Resource:
        """
        Register resource content for the given resource type name.

        Parameters
        ----------
        content : dict
        register_type : str

        Returns
        -------
        Resource

        """
        resource_type = get_object_or_404(ResourceType, slug=register_type)
        resource = Resource(content=content, resource_type=resource_type)
        resource = ResourceRegistration.register_instance(resource)
        return resource

    @staticmethod
    def register_instance(resource: Resource) -> Resource:
        """
        Register the resource for the given resource type.

        Parameters
        ----------
        resource : Resource
        resource_type : ResourceType

        Returns
        -------
        Resource

        """
        with transaction.atomic():
            resource.etag = uuid4().hex
            resource.modified_at = datetime.now()
            resource.resource_type.etag = resource.etag
            resource.resource_type.modified_at = resource.modified_at
            resource.save()
            resource.resource_type.save()
        return resource

    @staticmethod
    def update(content: str, register_type: str, slug: str) -> Resource:
        """
        Update resource content for the given resource type name and slug.

        Parameters
        ----------
        content : dict
        register_type : str
        slug : str

        Returns
        -------
        Resource

        """
        with transaction.atomic():
            resource_type = get_object_or_404(ResourceType, slug=register_type)

            # note: for now slug is used as ID
            resource = get_object_or_404(
                Resource, resource_type=resource_type, pk=int(slug)
            )

            resource.content = content
            resource.etag = uuid4().hex
            resource.modified_at = datetime.now()
            resource_type.etag = resource.etag
            resource_type.modified_at = resource.modified_at
            resource.save()
            resource_type.save()
        return resource

    @staticmethod
    def delete(register_type: str, slug: str):
        """
        Delete resource for the given resource type name and slug.

        Parameters
        ----------
        register_type : str
        slug : str

        """
        with transaction.atomic():
            resource_type = get_object_or_404(ResourceType, slug=register_type)

            # note: for now slug is used as ID
            resource = get_object_or_404(
                Resource, resource_type=resource_type, pk=int(slug)
            )
            resource.delete()
            resource_type.etag = uuid4().hex
            resource_type.modified_at = datetime.now()
            resource_type.save()


JSON_SCHEMA_REGISTRATION_ = None
JSON_SCHEMA_RESOURCE_TYPE_ = None
JSON_LD_RESOURCE_TYPE_ = None
JSON_LD_REGISTRATION_ = None


def register_json_schema_resource_type():
    global JSON_SCHEMA_RESOURCE_TYPE_
    global JSON_SCHEMA_REGISTRATION_
    if JSON_SCHEMA_RESOURCE_TYPE_ is None:
        JSON_SCHEMA_REGISTRATION_ = ResourceTypeRegistration(
            resource_type_uri="http://json-schema.org/draft-04/schema",
            revision="1",
            schema_uri="http://json-schema.org/draft-04/schema",
            slug="json-schema",
        )
        JSON_SCHEMA_RESOURCE_TYPE_ = JSON_SCHEMA_REGISTRATION_.resource_type
    return JSON_SCHEMA_RESOURCE_TYPE_


def register_json_ld_resource_type():
    global JSON_LD_RESOURCE_TYPE_
    global JSON_LD_REGISTRATION_
    if JSON_LD_RESOURCE_TYPE_ is None:
        JSON_LD_REGISTRATION_ = ResourceTypeRegistration(
            resource_type_uri="http://www.w3.org/ns/json-ld#context",
            revision="1",
            slug="json-ld",
        )
        JSON_LD_RESOURCE_TYPE_ = JSON_LD_REGISTRATION_.resource_type
    return JSON_LD_RESOURCE_TYPE_


def register_json_file(file_contents_func, previous_revision, resource_type):
    contents = file_contents_func()
    if type(contents) is str:
        contents = json.loads(contents)
    if previous_revision:
        resource = ResourceRegistration.update(
            contents, resource_type.slug, previous_revision.id
        )
    else:
        resource = ResourceRegistration.register(contents, resource_type.slug)
    return resource


def register_json_schema(schema_func, previous_revision):
    resource_type = register_json_schema_resource_type()
    return register_json_file(schema_func, previous_revision, resource_type)


def register_json_ld_context(context_func, previous_revision):
    resource_type = register_json_ld_resource_type()
    return register_json_file(context_func, previous_revision, resource_type)


def json_resource_type(
    resource_type_uri: str,
    revision: str,
    slug: str,
    schema_resource: Resource = None,
    schema_func=None,
    schema_uri: str = None,
    context_resource: Resource = None,
    context_func=None,
    context_uri: str = None,
):
    """Registration of a single json resource type. Callers **should** provide one and
    only one of `schema_resource, schema_path, schema_uri` as well as one and only  one
    of `context_resource, context_path, context_uri`.
    """
    try:
        resource_type_registration = ResourceTypeRegistration(
            resource_type_uri,
            revision,
            slug=slug,
            schema_resource=schema_resource,
            schema_func=schema_func,
            schema_uri=schema_uri,
            context_resource=context_resource,
            context_func=context_func,
            context_uri=context_uri,
        )
        return resource_type_registration.urls
    except (OperationalError, ProgrammingError):
        logging.warning(
            "Failed to register resource type due to OperationalError. "
            "This is expected for migrations."
        )
    return []


def resource_types(*args):
    """Convenience function to make the configuration of resource type registrations
    more readable.
    """
    ret = []
    for urls in args:
        ret.extend(urls)
    return ret



from rest_framework.exceptions import APIException
from rest_framework.views import exception_handler as drf_exception_handler


def scrud_exception_handler(exc, context):
    response = drf_exception_handler(exc, context)
    if isinstance(exc, ScrudException):
        response.data = exc.response_data(context)
        response.data["status"] = response.status_code
        try:
            request = context["request"]
            if request.accepted_renderer.format == "json":
                response.content_type = "application/problem+json"
        except AttributeError:
            pass
    return response


class ScrudException(APIException):
    def response_data(self, context):
        """Return RFC 7807 response data"""
        data = {}
        for property_name in ["type", "title", "detail", "instance"]:
            if hasattr(self, property_name):
                data[property_name] = getattr(self, property_name)
        return data


class JsonValidationError(ScrudException):
    status_code = 400
    default_detail = "Bad Request"

    def __init__(self, json_schema_exc):
        super().__init__()
        self.type = "https://api.openteams.com/http-problem/json-invalid"
        self.title = "JSON Validation Error"
        self.detail = json_schema_exc.message
        self.exc = json_schema_exc

    def response_data(self, context):
        data = super().response_data(context)
        data["name"] = self.exc.name
        data["value"] = self.exc.value
        data["definition"] = self.exc.definition
        data["rule"] = self.exc.rule
        data["rule_definition"] = self.exc.rule_definition
        return data



import fastjsonschema

from scrud_django.exceptions import JsonValidationError
from scrud_django.models import ResourceType


class ValidatorEntry:
    def __init__(self, resource_type: ResourceType, validate: callable):
        self.type_uri = resource_type.type_uri
        self.revision = resource_type.revision
        self.etag = resource_type.schema.etag
        self.modified_at = resource_type.schema.modified_at
        self.validate = validate

    def matches(self, resource_type: ResourceType):
        return (
            self.type_uri == resource_type.type_uri
            and self.revision == resource_type.revision
            and self.etag == resource_type.schema.etag
            and self.modified_at == resource_type.schema.modified_at
        )


class JsonValidator:
    requires_context = True

    def __init__(self):
        self.validators = {}

    def __call__(self, data, context):
        validate = self.validator_for(context.resource_type)
        data = validate(data)
        return data

    def is_json_schema(self, schema):
        return schema is not None and schema.resource_type.type_uri in [
            "http://json-schema.org/draft-04/schema",
            "http://json-schema.org/draft-06/schema",
            "http://json-schema.org/draft-07/schema",
        ]

    def validator_for(self, resource_type):
        schema = resource_type.schema
        if not self.is_json_schema(schema):
            return lambda r: r
        validator = self.validators.get(resource_type.type_uri, None)
        if validator is None or not validator.matches(resource_type):
            try:
                content = schema.content
                if "$id" in content:
                    del content["$id"]
                validation_func = fastjsonschema.compile(content)
            except Exception as e:
                import logging

                logging.error(e)
                return lambda r: r

            def validate(content, handlers={}, formats={}):
                try:
                    return validation_func(content)
                except fastjsonschema.JsonSchemaException as e:
                    raise JsonValidationError(e)

            validator = ValidatorEntry(resource_type, validate)
            self.validators[resource_type.type_uri] = validator
        return validator.validate



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scrud_django', 'scrud_django.migrations']

package_data = \
{'': ['*']}

install_requires = \
['django-cors-headers>=3.6.0,<4.0.0',
 'django-jsonfield>=1.4.1,<2.0.0',
 'django-scoped-rbac==0.2.2',
 'django>=2.2.8,<3.0.0',
 'djangorestframework>=3.11.0,<4.0.0',
 'fastjsonschema>=2.14.5,<3.0.0',
 'psycopg2>=2.8.6,<3.0.0',
 'python-decouple>=3.3,<4.0',
 'python-dotenv>=0.15.0,<0.16.0']

setup_kwargs = {
    'name': 'scrud-django',
    'version': '0.2.1',
    'description': 'A Django REST Framework app for Semantic CRUD.',
    'long_description': None,
    'author': 'David Charboneau',
    'author_email': 'david@openteams.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
